
#include <stdlib.h>
#include <stdio.h>

#define NAGRA
#define TESTER
#include "crypto.h"
#include "data.h"
#include "system-common.h"
#include "system-nagra.c"

#include "compat.h"

class cNagraDESold {
private:
  static const unsigned char SCODES[];
  //
  void Crypt(const unsigned char *Data, const unsigned char *Key, unsigned char *Out);
  void Hash(const unsigned char *data, unsigned char *hash, const int rounds);
  unsigned char Mirrorbits(unsigned char b);
  unsigned char BitOf7(const unsigned char *b, int nb);
  void PrepareKey(const unsigned char *inKey, unsigned char *outKey);
  void InitializeBuffer(unsigned char *outBuff);
  void PrepareKey2(const unsigned char *inKey, unsigned char *outBuff);
  void Permute_L(unsigned char *b);
  void Permute_R(const unsigned char *inKey, unsigned char *outBuff);
  void StripParity(const unsigned char *in, unsigned char *out);
  void PrepareSBox(unsigned char *parity, unsigned char *buffer, unsigned char *cmd4);
  void CryptStep2(const unsigned char *inData, unsigned char *outBuff);
  void DecryptStep2(const unsigned char *inData, unsigned char *outBuff);
  void MoveBuffer(const unsigned char *inBuff, unsigned char *outData);
  void PermutationDecrypt(unsigned char *buffer, unsigned char *data, unsigned char *parity, unsigned char *cmd4);
  void PermutationCrypt(unsigned char *buffer, unsigned char *data, unsigned char *parity, unsigned char *cmd4);
  void Permutation2(const unsigned char *parity, unsigned char *data);
protected:
  bool SigCheck(const unsigned char *block, const unsigned char *sig, const unsigned char *vkey, const int rounds);
  void DecryptBlock(const unsigned char *crypted, const unsigned char *key, unsigned char *decrypted, const int rounds);
  //
  void Decrypt(const unsigned char *data, const unsigned char *key, unsigned char *decrypted, bool desmod=false);
  void RotateBytes(unsigned char *in, int n);
};

const unsigned char cNagraDESold::SCODES[256] = {
  0xF7, 0xE2, 0x3C, 0xBF, 0x04, 0x5B, 0x9A, 0xEC, 0x9B, 0x87, 0x06, 0x59,
  0xAD, 0xF4, 0xE0, 0x0A, 0x62, 0x08, 0x95, 0x63, 0x5F, 0x36, 0x49, 0x95,
  0x38, 0xD1, 0x73, 0xAE, 0x91, 0xAD, 0x8E, 0xF0, 0x10, 0x7F, 0x55, 0x1A,
  0xE7, 0x92, 0xF9, 0x25, 0x4E, 0x11, 0xC3, 0xCC, 0x3B, 0xC8, 0x2C, 0x56,
  0xCF, 0x93, 0x66, 0xDD, 0xB4, 0xE9, 0x3A, 0x70, 0xA2, 0x24, 0xAD, 0x37,
  0x48, 0x1E, 0xD1, 0xAB, 0x85, 0xBB, 0x88, 0x4D, 0x76, 0x61, 0x7D, 0xDA,
  0xE9, 0x42, 0xB3, 0x94, 0x1F, 0x8C, 0x04, 0x67, 0xD0, 0x66, 0x4B, 0xF8,
  0x2C, 0xDF, 0x12, 0x35, 0x07, 0x39, 0xEE, 0x03, 0xCA, 0x40, 0xB1, 0xCE,
  0x7B, 0xC8, 0xF4, 0x22, 0xDC, 0x06, 0xA3, 0x8D, 0xB0, 0xAB, 0x2A, 0x77,
  0x66, 0x71, 0x5F, 0xB4, 0x2E, 0x55, 0x11, 0x8F, 0x82, 0xB9, 0xCD, 0x4A,
  0x59, 0xF0, 0xD7, 0xEC, 0xF5, 0x2E, 0x68, 0x13, 0xC4, 0xB2, 0x51, 0x8F,
  0xBE, 0xC5, 0x2B, 0x46, 0x32, 0x28, 0x6C, 0x33, 0xDD, 0x9E, 0xD7, 0xF0,
  0xF3, 0x64, 0xEA, 0x59, 0xC5, 0x5B, 0x90, 0x2C, 0x68, 0x8D, 0x0F, 0x0A,
  0x06, 0x31, 0x69, 0xC7, 0x27, 0xDD, 0x2A, 0xB6, 0x52, 0x08, 0x4C, 0x75,
  0x84, 0x43, 0xBF, 0x60, 0xEB, 0xA4, 0x81, 0x9A, 0x1D, 0x01, 0x90, 0xEF,
  0x2E, 0xB7, 0xF9, 0x12, 0xD3, 0x7E, 0xC5, 0x99, 0x78, 0x4B, 0x16, 0x6C,
  0xB2, 0x18, 0xFC, 0xF5, 0x1F, 0xF3, 0xCA, 0x80, 0x04, 0xED, 0x89, 0xA6,
  0x61, 0x2E, 0x76, 0x39, 0x4D, 0xF2, 0x33, 0xCF, 0xF0, 0x8C, 0xA5, 0x5A,
  0x97, 0x5B, 0xDE, 0x71, 0xAB, 0xE7, 0x08, 0xA4, 0xEB, 0xA6, 0x47, 0x19,
  0x82, 0x68, 0x34, 0xE7, 0x5D, 0x3B, 0x7A, 0xD0, 0x38, 0xD5, 0xE1, 0x0C,
  0x70, 0xCD, 0xAC, 0x2A, 0x49, 0x12, 0x5F, 0xB4, 0xAE, 0x91, 0x13, 0x4F,
  0x95, 0x7E, 0xB6, 0xD3
};

void cNagraDESold::RotateBytes(unsigned char *in, int n)
{
    int j=n-1;
    n /= 2;
    for(int i=0; i<n ; i++,j--) {
       unsigned char temp = in[i];
       in[i] = in[j];
       in[j] = temp;
       }
}

unsigned char cNagraDESold::Mirrorbits(unsigned char b)
{
  unsigned char c=0;
  for(int i=0; i<8; i++, b=b/2) c=2*c+b%2;
  return c;
}

unsigned char cNagraDESold::BitOf7(const unsigned char *b, int nb)
{
  return (b[6 - (nb / 8)] >> (nb % 8)) & 1;
}

void cNagraDESold::PrepareKey(const unsigned char *inKey, unsigned char *outKey)
{
  for(int i=0; i<8; i++) outKey[i]=2*Mirrorbits(inKey[i]);
  // Compress Nagra_KEY
                   outKey[0] |= outKey[1] >> 7;
  outKey[1] <<= 1; outKey[1] |= outKey[2] >> 6;
  outKey[2] <<= 2; outKey[2] |= outKey[3] >> 5;
  outKey[3] <<= 3; outKey[3] |= outKey[4] >> 4;
  outKey[4] <<= 4; outKey[4] |= outKey[5] >> 3;
  outKey[5] <<= 5; outKey[5] |= outKey[6] >> 2;
  outKey[6] <<= 6; outKey[6] |= outKey[7] >> 1;
  outKey[7] = 0;
}

void cNagraDESold::InitializeBuffer(unsigned char *outBuff)
{
  outBuff[0] = 0x0;
  outBuff[1] = 0;
  outBuff[2] = 0x40;
  outBuff[3] = 0x01;
  outBuff[4] = 0x80;
  outBuff[5] = 0x02;
  outBuff[6] = 0xC0;
  outBuff[7] = 0x03;
  outBuff[8] = 0x0;
}

void cNagraDESold::PrepareKey2(const unsigned char *inKey, unsigned char *outBuff)
{
  outBuff[0] |= (BitOf7(inKey, 21) << 5) + (BitOf7(inKey, 49) << 4) + (BitOf7(inKey, 2) << 3) + (BitOf7(inKey, 36) << 2) + (BitOf7(inKey, 51) << 1) + BitOf7(inKey, 15);
  outBuff[1] |= (BitOf7(inKey, 43) << 7) + (BitOf7(inKey, 23) << 6) + (BitOf7(inKey, 14) << 5) + (BitOf7(inKey, 8) << 4) + (BitOf7(inKey, 31) << 3) + (BitOf7(inKey, 35) << 2);
  outBuff[2] |= (BitOf7(inKey,  0) << 5) + (BitOf7(inKey, 45) << 4) + (BitOf7(inKey, 28) << 3) + (BitOf7(inKey, 29) << 2) + (BitOf7(inKey, 37) << 1) + BitOf7(inKey, 9);
  outBuff[3] |= (BitOf7(inKey, 42) << 7) + (BitOf7(inKey, 22) << 6) + (BitOf7(inKey, 30) << 5) + (BitOf7(inKey, 38) << 4) + (BitOf7(inKey, 7) << 3) + (BitOf7(inKey, 1) << 2);
  outBuff[4] |= (BitOf7(inKey, 10) << 5) + (BitOf7(inKey, 39) << 4) + (BitOf7(inKey, 54) << 3) + (BitOf7(inKey, 41) << 2) + (BitOf7(inKey, 4) << 1) + BitOf7(inKey, 26);
  outBuff[5] |= (BitOf7(inKey, 32) << 7) + (BitOf7(inKey, 27) << 6) + (BitOf7(inKey, 53) << 5) + (BitOf7(inKey, 11) << 4) + (BitOf7(inKey, 33) << 3) + (BitOf7(inKey, 48) << 2);
  outBuff[6] |= (BitOf7(inKey, 24) << 5) + (BitOf7(inKey, 20) << 4) + (BitOf7(inKey, 3) << 3) + (BitOf7(inKey, 40) << 2) + (BitOf7(inKey, 25) << 1) + BitOf7(inKey, 5);
  outBuff[7] |= (BitOf7(inKey, 34) << 7) + (BitOf7(inKey, 55) << 6) + (BitOf7(inKey, 6) << 5) + (BitOf7(inKey, 18) << 4) + (BitOf7(inKey, 19) << 3) + (BitOf7(inKey, 46) << 2);
  outBuff[8] |= (BitOf7(inKey, 17) << 7) + (BitOf7(inKey, 12) << 6) + (BitOf7(inKey, 47) << 5) + (BitOf7(inKey, 13) << 4) + (BitOf7(inKey, 52) << 3) + (BitOf7(inKey, 16) << 2) + (BitOf7(inKey, 44) << 1) + BitOf7(inKey, 50);
}

void cNagraDESold::Permute_L(unsigned char *b)
{
  for(int i=0; i<8; i++) b[i]=Mirrorbits(b[i]);
}

void cNagraDESold::Permute_R(const unsigned char *inKey, unsigned char *outBuff)
{
  if (inKey[0] & 1)   outBuff[7] |= (1 << 6);
  if (inKey[0] & 2)   outBuff[1] |= (1 << 3);
  if (inKey[0] & 4)   outBuff[2] |= (1 << 5);
  if (inKey[0] & 8)   outBuff[3] |= (1 << 2);
  if (inKey[0] & 16)  outBuff[0] |= (1 << 3);
  if (inKey[0] & 32)  outBuff[6] |= (1 << 0);
  if (inKey[0] & 64)  outBuff[7] |= (1 << 5);
  if (inKey[0] & 128) outBuff[6] |= (1 << 3);

  if (inKey[1] & 1)   outBuff[8] |= (1 << 5);
  if (inKey[1] & 2)   outBuff[5] |= (1 << 2);
  if (inKey[1] & 4)   outBuff[0] |= (1 << 4);
  if (inKey[1] & 8)   outBuff[8] |= (1 << 0);
  if (inKey[1] & 16)  outBuff[0] |= (1 << 1);
  if (inKey[1] & 32)  outBuff[8] |= (1 << 3);
  if (inKey[1] & 64)  outBuff[5] |= (1 << 5);
  if (inKey[1] & 128) outBuff[4] |= (1 << 3);

  if (inKey[2] & 1)   outBuff[4] |= (1 << 4);
  if (inKey[2] & 2)   outBuff[6] |= (1 << 2);
  if (inKey[2] & 4)   outBuff[4] |= (1 << 2);
  if (inKey[2] & 8)   outBuff[3] |= (1 << 7);
  if (inKey[2] & 16)  outBuff[1] |= (1 << 7);
  if (inKey[2] & 32)  outBuff[8] |= (1 << 1);
  if (inKey[2] & 64)  outBuff[2] |= (1 << 4);
  if (inKey[2] & 128) outBuff[7] |= (1 << 2);

  if (inKey[3] & 1)   outBuff[4] |= (1 << 1);
  if (inKey[3] & 2)   outBuff[5] |= (1 << 7);
  if (inKey[3] & 4)   outBuff[5] |= (1 << 3);
  if (inKey[3] & 8)   outBuff[7] |= (1 << 7);
  if (inKey[3] & 16)  outBuff[1] |= (1 << 2);
  if (inKey[3] & 32)  outBuff[0] |= (1 << 2);
  if (inKey[3] & 64)  outBuff[2] |= (1 << 1);
  if (inKey[3] & 128) outBuff[3] |= (1 << 4);

  if (inKey[4] & 1)   outBuff[1] |= (1 << 6);
  if (inKey[4] & 2)   outBuff[6] |= (1 << 5);
  if (inKey[4] & 4)   outBuff[6] |= (1 << 1);
  if (inKey[4] & 8)   outBuff[4] |= (1 << 0);
  if (inKey[4] & 16)  outBuff[5] |= (1 << 6);
  if (inKey[4] & 32)  outBuff[2] |= (1 << 3);
  if (inKey[4] & 64)  outBuff[2] |= (1 << 2);
  if (inKey[4] & 128) outBuff[3] |= (1 << 5);

  if (inKey[5] & 1)   outBuff[0] |= (1 << 0);
  if (inKey[5] & 2)   outBuff[8] |= (1 << 2);
  if (inKey[5] & 4)   outBuff[8] |= (1 << 7);
  if (inKey[5] & 8)   outBuff[7] |= (1 << 4);
  if (inKey[5] & 16)  outBuff[7] |= (1 << 3);
  if (inKey[5] & 32)  outBuff[6] |= (1 << 4);
  if (inKey[5] & 64)  outBuff[0] |= (1 << 5);
  if (inKey[5] & 128) outBuff[3] |= (1 << 6);

  if (inKey[6] & 1)   outBuff[3] |= (1 << 3);
  if (inKey[6] & 2)   outBuff[1] |= (1 << 4);
  if (inKey[6] & 4)   outBuff[2] |= (1 << 0);
  if (inKey[6] & 8)   outBuff[4] |= (1 << 5);
  if (inKey[6] & 16)  outBuff[5] |= (1 << 4);
  if (inKey[6] & 32)  outBuff[8] |= (1 << 6);
  if (inKey[6] & 64)  outBuff[8] |= (1 << 4);
  if (inKey[6] & 128) outBuff[1] |= (1 << 5);
}

void cNagraDESold::StripParity(const unsigned char *in, unsigned char *out)
{
  for (int i = 0; i < 8; i++) {
    out[i] = 0;
    for (int b = 0; b < 8; b++) {
      if ((in[b] >> (7 - i)) & 1)
        out[i] |= (1 << b);
      }
    }
}

void cNagraDESold::PrepareSBox(unsigned char *parity, unsigned char *buffer, unsigned char *cmd4)
{
  unsigned char temp;
  unsigned char databyte;
  cmd4[0] = 0;
  cmd4[1] = 0;
  cmd4[2] = 0;
  cmd4[3] = 0;
  // 0
  temp = (parity[7] * 2) + (parity[1] >> 7);
  databyte = SCODES[(temp & 0x3F) ^ buffer[0]]; // SBox_R
  if (databyte & 1)   cmd4[1] |= (1 << 0);
  if (databyte & 2)   cmd4[2] |= (1 << 0);
  if (databyte & 4)   cmd4[2] |= (1 << 6);
  if (databyte & 8)   cmd4[3] |= (1 << 6);
  // 1
  temp = (parity[7] / 2) + ((parity[5] & 1) << 7);
  databyte = SCODES[(temp & 0xFC) ^ buffer[1]]; // SBox_L
  if (databyte & 16)  cmd4[1] |= (1 << 4);
  if (databyte & 32)  cmd4[3] |= (1 << 3);
  if (databyte & 64)  cmd4[0] |= (1 << 1);
  if (databyte & 128) cmd4[2] |= (1 << 1);
  // 2
  temp = (parity[5] * 2) + (parity[7] >> 7);
  databyte = SCODES[(temp & 0x3F) ^ buffer[2]]; // SBox_R
  if (databyte & 1)   cmd4[2] |= (1 << 7);
  if (databyte & 2)   cmd4[1] |= (1 << 7);
  if (databyte & 4)   cmd4[3] |= (1 << 5);
  if (databyte & 8)   cmd4[0] |= (1 << 5);
  // 3
  temp = (parity[5] / 2) + ((parity[3] & 1) << 7);
  databyte = SCODES[(temp & 0xFC) ^ buffer[3]]; // SBox_L
  if (databyte & 16)  cmd4[3] |= (1 << 1);
  if (databyte & 32)  cmd4[2] |= (1 << 3);
  if (databyte & 64)  cmd4[1] |= (1 << 1);
  if (databyte & 128) cmd4[0] |= (1 << 0);
  // 4
  temp = (parity[3] * 2) + (parity[5] >> 7);
  databyte = SCODES[(temp & 0x3F) ^ buffer[4]]; // SBox_R
  if (databyte & 1)   cmd4[0] |= (1 << 7);
  if (databyte & 2)   cmd4[1] |= (1 << 5);
  if (databyte & 4)   cmd4[3] |= (1 << 0);
  if (databyte & 8)   cmd4[0] |= (1 << 2);
  // 5
  temp = (parity[3] / 2) + ((parity[1] & 1) << 7);
  databyte = SCODES[(temp & 0xFC) ^ buffer[5]]; // SBox_L
  if (databyte & 16)  cmd4[0] |= (1 << 3);
  if (databyte & 32)  cmd4[3] |= (1 << 4);
  if (databyte & 64)  cmd4[1] |= (1 << 2);
  if (databyte & 128) cmd4[2] |= (1 << 2);
  // 6
  temp = (parity[1] * 2) + (parity[3] >> 7);
  databyte = SCODES[(temp & 0x3F) ^ buffer[6]]; // SBox_R
  if (databyte & 1)   cmd4[3] |= (1 << 7);
  if (databyte & 2)   cmd4[1] |= (1 << 3);
  if (databyte & 4)   cmd4[2] |= (1 << 5);
  if (databyte & 8)   cmd4[0] |= (1 << 6);
  // 7
  temp = (parity[1] / 2) + ((parity[7] & 1) << 7);
  databyte = SCODES[(temp & 0xFC) ^ buffer[7]]; // SBox_L
  if (databyte & 16)  cmd4[0] |= (1 << 4);
  if (databyte & 32)  cmd4[3] |= (1 << 2);
  if (databyte & 64)  cmd4[1] |= (1 << 6);
  if (databyte & 128) cmd4[2] |= (1 << 4);
  //
  temp = parity[6];
  temp = temp ^ cmd4[0];
  parity[6] = parity[7];
  parity[7] = temp;

  temp = parity[4];
  temp = temp ^ cmd4[1];
  parity[4] = parity[5];
  parity[5] = temp;

  temp = parity[2];
  temp = temp ^ cmd4[2];
  parity[2] = parity[3];
  parity[3] = temp;

  temp = parity[0];
  temp = temp ^ cmd4[3];
  parity[0] = parity[1];
  parity[1] = temp;
}

void cNagraDESold::CryptStep2(const unsigned char *inData, unsigned char *outBuff)
{
  if (inData[0] & 1)   outBuff[3] |= (1 << 6);
  if (inData[0] & 2)   outBuff[3] |= (1 << 2);
  if (inData[0] & 4)   outBuff[1] |= (1 << 7);
  if (inData[0] & 8)   outBuff[2] |= (1 << 0);
  if (inData[0] & 16)  outBuff[1] |= (1 << 3);
  if (inData[0] & 32)  outBuff[2] |= (1 << 3);

  if (inData[1] & 4)   outBuff[3] |= (1 << 7);
  if (inData[1] & 8)   outBuff[3] |= (1 << 4);
  if (inData[1] & 16)  outBuff[0] |= (1 << 0);
  if (inData[1] & 32)  outBuff[0] |= (1 << 5);
  if (inData[1] & 64)  outBuff[3] |= (1 << 5);
  if (inData[1] & 128) outBuff[8] |= (1 << 0);

  if (inData[2] & 1)   outBuff[8] |= (1 << 2);
  if (inData[2] & 2)   outBuff[8] |= (1 << 1);
  if (inData[2] & 4)   outBuff[0] |= (1 << 2);
  if (inData[2] & 8)   outBuff[1] |= (1 << 2);
  if (inData[2] & 16)  outBuff[8] |= (1 << 3);
  if (inData[2] & 32)  outBuff[3] |= (1 << 3);

  if (inData[3] & 4)   outBuff[1] |= (1 << 4);
  if (inData[3] & 8)   outBuff[1] |= (1 << 5);
  if (inData[3] & 16)  outBuff[2] |= (1 << 4);
  if (inData[3] & 32)  outBuff[2] |= (1 << 1);
  if (inData[3] & 64)  outBuff[2] |= (1 << 2);
  if (inData[3] & 128) outBuff[0] |= (1 << 4);

  if (inData[4] & 1)   outBuff[5] |= (1 << 3);
  if (inData[4] & 2)   outBuff[5] |= (1 << 4);
  if (inData[4] & 4)   outBuff[5] |= (1 << 2);
  if (inData[4] & 8)   outBuff[7] |= (1 << 5);
  if (inData[4] & 16)  outBuff[7] |= (1 << 2);
  if (inData[4] & 32)  outBuff[8] |= (1 << 7);

  if (inData[5] & 4)   outBuff[7] |= (1 << 6);
  if (inData[5] & 8)   outBuff[6] |= (1 << 2);
  if (inData[5] & 16)  outBuff[7] |= (1 << 4);
  if (inData[5] & 32)  outBuff[6] |= (1 << 0);
  if (inData[5] & 64)  outBuff[7] |= (1 << 7);
  if (inData[5] & 128) outBuff[4] |= (1 << 4);

  if (inData[6] & 1)   outBuff[8] |= (1 << 6);
  if (inData[6] & 2)   outBuff[5] |= (1 << 7);
  if (inData[6] & 4)   outBuff[8] |= (1 << 5);
  if (inData[6] & 8)   outBuff[4] |= (1 << 5);
  if (inData[6] & 16)  outBuff[5] |= (1 << 6);
  if (inData[6] & 32)  outBuff[4] |= (1 << 1);

  if (inData[7] & 4)   outBuff[5] |= (1 << 5);
  if (inData[7] & 8)   outBuff[4] |= (1 << 0);
  if (inData[7] & 16)  outBuff[6] |= (1 << 1);
  if (inData[7] & 32)  outBuff[8] |= (1 << 4);
  if (inData[7] & 64)  outBuff[6] |= (1 << 3);
  if (inData[7] & 128) outBuff[4] |= (1 << 2);

  if (inData[8] & 1)   outBuff[2] |= (1 << 5);
  if (inData[8] & 2)   outBuff[0] |= (1 << 1);
  if (inData[8] & 4)   outBuff[1] |= (1 << 6);
  if (inData[8] & 8)   outBuff[0] |= (1 << 3);
  if (inData[8] & 16)  outBuff[6] |= (1 << 4);
  if (inData[8] & 32)  outBuff[4] |= (1 << 3);
  if (inData[8] & 64)  outBuff[7] |= (1 << 3);
  if (inData[8] & 128) outBuff[6] |= (1 << 5);
}

void cNagraDESold::DecryptStep2(const unsigned char *inData, unsigned char *outBuff)
{
  if (inData[0] & 1)   outBuff[1] |= (1 << 4);
  if (inData[0] & 2)   outBuff[8] |= (1 << 1);
  if (inData[0] & 4)   outBuff[2] |= (1 << 2);
  if (inData[0] & 8)   outBuff[8] |= (1 << 3);
  if (inData[0] & 16)  outBuff[3] |= (1 << 7);
  if (inData[0] & 32)  outBuff[1] |= (1 << 5);

  if (inData[1] & 4)   outBuff[2] |= (1 << 3);
  if (inData[1] & 8)   outBuff[0] |= (1 << 4);
  if (inData[1] & 16)  outBuff[3] |= (1 << 2);
  if (inData[1] & 32)  outBuff[3] |= (1 << 3);
  if (inData[1] & 64)  outBuff[8] |= (1 << 2);
  if (inData[1] & 128) outBuff[0] |= (1 << 2);

  if (inData[2] & 1)   outBuff[0] |= (1 << 3);
  if (inData[2] & 2)   outBuff[3] |= (1 << 5);
  if (inData[2] & 4)   outBuff[3] |= (1 << 6);
  if (inData[2] & 8)   outBuff[0] |= (1 << 5);
  if (inData[2] & 16)  outBuff[3] |= (1 << 4);
  if (inData[2] & 32)  outBuff[8] |= (1 << 0);

  if (inData[3] & 4)   outBuff[0] |= (1 << 1);
  if (inData[3] & 8)   outBuff[2] |= (1 << 5);
  if (inData[3] & 16)  outBuff[1] |= (1 << 3);
  if (inData[3] & 32)  outBuff[1] |= (1 << 6);
  if (inData[3] & 64)  outBuff[0] |= (1 << 0);
  if (inData[3] & 128) outBuff[1] |= (1 << 2);

  if (inData[4] & 1)   outBuff[7] |= (1 << 3);
  if (inData[4] & 2)   outBuff[6] |= (1 << 5);
  if (inData[4] & 4)   outBuff[7] |= (1 << 7);
  if (inData[4] & 8)   outBuff[8] |= (1 << 5);
  if (inData[4] & 16)  outBuff[5] |= (1 << 7);
  if (inData[4] & 32)  outBuff[6] |= (1 << 3);

  if (inData[5] & 4)   outBuff[4] |= (1 << 2);
  if (inData[5] & 8)   outBuff[4] |= (1 << 0);
  if (inData[5] & 16)  outBuff[4] |= (1 << 1);
  if (inData[5] & 32)  outBuff[7] |= (1 << 2);
  if (inData[5] & 64)  outBuff[6] |= (1 << 4);
  if (inData[5] & 128) outBuff[6] |= (1 << 1);

  if (inData[6] & 1)   outBuff[5] |= (1 << 5);
  if (inData[6] & 2)   outBuff[7] |= (1 << 4);
  if (inData[6] & 4)   outBuff[5] |= (1 << 3);
  if (inData[6] & 8)   outBuff[7] |= (1 << 6);
  if (inData[6] & 16)  outBuff[8] |= (1 << 4);
  if (inData[6] & 32)  outBuff[8] |= (1 << 7);

  if (inData[7] & 4)   outBuff[4] |= (1 << 4);
  if (inData[7] & 8)   outBuff[8] |= (1 << 6);
  if (inData[7] & 16)  outBuff[5] |= (1 << 4);
  if (inData[7] & 32)  outBuff[4] |= (1 << 3);
  if (inData[7] & 64)  outBuff[5] |= (1 << 2);
  if (inData[7] & 128) outBuff[5] |= (1 << 6);

  if (inData[8] & 1)   outBuff[1] |= (1 << 7);
  if (inData[8] & 2)   outBuff[2] |= (1 << 1);
  if (inData[8] & 4)   outBuff[2] |= (1 << 0);
  if (inData[8] & 8)   outBuff[2] |= (1 << 4);
  if (inData[8] & 16)  outBuff[7] |= (1 << 5);
  if (inData[8] & 32)  outBuff[6] |= (1 << 2);
  if (inData[8] & 64)  outBuff[6] |= (1 << 0);
  if (inData[8] & 128) outBuff[4] |= (1 << 5);
}

void cNagraDESold::MoveBuffer(const unsigned char *inBuff, unsigned char *outData)
{
  memcpy(outData, inBuff, 9);
}

void cNagraDESold::PermutationDecrypt(unsigned char *buffer, unsigned char *data, unsigned char *parity, unsigned char *cmd4)
{
  for (int loop = 0; loop < 6; loop++) {
    InitializeBuffer(buffer);
    DecryptStep2(data, buffer);
    PrepareSBox(parity, buffer, cmd4);
    MoveBuffer(buffer, data);
    InitializeBuffer(buffer);
    DecryptStep2(data, buffer);
    MoveBuffer(buffer, data);
    }
}

void cNagraDESold::PermutationCrypt(unsigned char *buffer, unsigned char *data, unsigned char *parity, unsigned char *cmd4)
{
  for (int loop = 0; loop < 6; loop++) {
    InitializeBuffer(buffer);
    CryptStep2(data, buffer);
    PrepareSBox(parity, buffer, cmd4);
    MoveBuffer(buffer, data);
    InitializeBuffer(buffer);
    CryptStep2(data, buffer);
    MoveBuffer(buffer, data);
    }
}

void cNagraDESold::Permutation2(const unsigned char *parity, unsigned char *data)
{
  data[8] = 1;
  for (int loop = 0; loop < 8; loop++) {
    data[loop] = 0;
    if (parity[1] & data[8]) data[loop] |= (1 << 7);
    if (parity[0] & data[8]) data[loop] |= (1 << 6);
    if (parity[3] & data[8]) data[loop] |= (1 << 5);
    if (parity[2] & data[8]) data[loop] |= (1 << 4);
    if (parity[5] & data[8]) data[loop] |= (1 << 3);
    if (parity[4] & data[8]) data[loop] |= (1 << 2);
    if (parity[7] & data[8]) data[loop] |= (1 << 1);
    if (parity[6] & data[8]) data[loop] |= (1 << 0);
    data[8] <<= 1;
    }
}

void cNagraDESold::Decrypt(const unsigned char *_data, const unsigned char *_key, unsigned char *_decrypted, bool desmod)
{
  unsigned char data[9];
  unsigned char dataparity[8];
  unsigned char cardkey[8];
  unsigned char nagrakey[8];
  unsigned char nagrabuffer[9];
  unsigned char nagracmd4[4];
  memcpy(data, _data, 8);
  data[8] = 0;
  memcpy(cardkey, _key, 8);

// Nagra Algo
  PrepareKey(cardkey, nagrakey);
  InitializeBuffer(nagrabuffer);
  PrepareKey2(nagrakey, nagrabuffer);
  if(desmod) RotateBytes(data, 8); // for P & Q Calculation
  Permute_L(data);
  StripParity(data, dataparity);
  PrepareSBox(dataparity, nagrabuffer, nagracmd4);
  MoveBuffer(nagrabuffer, data);
  PermutationDecrypt(nagrabuffer, data, dataparity, nagracmd4);
  InitializeBuffer(nagrabuffer);
  DecryptStep2(data, nagrabuffer);
  MoveBuffer(nagrabuffer, data);
  PrepareSBox(dataparity, nagrabuffer, nagracmd4);
  PermutationDecrypt(nagrabuffer, data, dataparity, nagracmd4);
  InitializeBuffer(nagrabuffer);
  DecryptStep2(data, nagrabuffer);
  PrepareSBox(dataparity, nagrabuffer, nagracmd4);
  MoveBuffer(nagrabuffer, data);
  InitializeBuffer(nagrabuffer);
  DecryptStep2(data, nagrabuffer);
  PrepareSBox(dataparity, nagrabuffer, nagracmd4);

  Permutation2(dataparity, data);
  if(desmod) RotateBytes(data, 8); // for P & Q Calculation
  Permute_L(data);
// End Nagra Algo
  memcpy(_decrypted, data, 8);
}

void cNagraDESold::Crypt(const unsigned char *Data, const unsigned char *Key, unsigned char *Out)
{
  unsigned char data[9];
  unsigned char dataparity[8];
  unsigned char cardkey[8];
  unsigned char nagrakey[8];
  unsigned char nagrabuffer[9];
  unsigned char nagracmd4[4];
  
  memcpy(data, Data, 8);
  data[8] = 0;
  memcpy(cardkey, Key, 8);
// Nagra Algo
  PrepareKey(cardkey, nagrakey);
  InitializeBuffer(nagrabuffer);
  Permute_R(nagrakey, nagrabuffer);
  Permute_L(data);
  StripParity(data, dataparity);
  PrepareSBox(dataparity, nagrabuffer, nagracmd4);
  MoveBuffer(nagrabuffer, data);
  PermutationCrypt(nagrabuffer, data, dataparity, nagracmd4);
  InitializeBuffer(nagrabuffer);
  CryptStep2(data, nagrabuffer);
  MoveBuffer(nagrabuffer, data);
  PrepareSBox(dataparity, nagrabuffer, nagracmd4);
  PermutationCrypt(nagrabuffer, data, dataparity, nagracmd4);
  InitializeBuffer(nagrabuffer);
  CryptStep2(data, nagrabuffer);
  PrepareSBox(dataparity, nagrabuffer, nagracmd4);
  MoveBuffer(nagrabuffer, data);
  InitializeBuffer(nagrabuffer);
  CryptStep2(data, nagrabuffer);
  PrepareSBox(dataparity, nagrabuffer, nagracmd4);
  Permutation2(dataparity, data);
  Permute_L(data);
// End Nagra Algo
  memcpy(Out, data, 8);
}

void cNagraDESold::Hash(const unsigned char *data, unsigned char *hash, const int rounds)
{
  unsigned char cr[8];
  for (int j = 0; j < rounds; j++) {
    Crypt(data+j*8, hash, cr);
    for (int i = 0; i < 8; i++) hash[i] = cr[i]^data[j*8+i];
    }
}

void cNagraDESold::DecryptBlock(const unsigned char *crypted, const unsigned char *key, unsigned char *decrypted, const int rounds)
{
  for (int i = 0; i < rounds; i++) Decrypt(crypted+i*8, key, decrypted+i*8);
}

bool cNagraDESold::SigCheck(const unsigned char *block, const unsigned char *sig, const unsigned char *vkey, const int rounds)
{
  unsigned char hash[8];
  memcpy(hash, vkey, 8);
  Hash(block, hash, rounds);
  return (0 == memcmp(hash, sig, 8));
}

class cNagraOld : public cNagraDESold {
public:
  void Test(const unsigned char *Data, const unsigned char *Key, unsigned char *Out);
  };

void cNagraOld::Test(const unsigned char *Data, const unsigned char *Key, unsigned char *Out)
{
  Decrypt(Data,Key,Out);
}

class cNagraNew : public cNagraDES {
public:
  void Test(const unsigned char *Data, const unsigned char *Key, unsigned char *Out);
  };

void cNagraNew::Test(const unsigned char *Data, const unsigned char *Key, unsigned char *Out)
{
  Decrypt(Data,Key,Out);
}

int main(int argc, char *argv[])
{
  if(argc<2) {
    printf("usage: %s <test-count>\n",argv[0]);
    return 1;
    }

  srandom(time(0));
  int num=atoi(argv[1]);
  for(int i=num; i>0 ; i--)
    {
    unsigned char data[8], key[8], out1[8], out2[8];
    for(int l=0 ; l<8; l++) {
      data[l]=random();
      key[l]=random();
      }
    cNagraNew nn;
    nn.Test(data,key,out2);
    cNagraOld on;
    on.Test(data,key,out1);
    if(memcmp(out1,out2,8)) {
      printf("Failed with:\n");
      printf("data    = "); SDump(data,8);
      printf("key     = "); SDump(key,8);
      printf("out-old = "); SDump(out1,8);
      printf("out-new = "); SDump(out2,8);
      break;
      }
    }
  printf("%d tests passed\n",num);

  unsigned char data[8], key[8], out[8];
  for(int l=0 ; l<8; l++) {
    data[l]=random();
    key[l]=random();
    }
  int t=time_ms();
  for(int i=num; i>0 ; i--)
    {
    cNagraOld n;
    n.Test(data,key,out);
    }
  t=time_ms()-t;
  printf("%d old speed tests: %f s\n",num,(float)t/1000.0);

  t=time_ms();
  for(int i=num; i>0 ; i--)
    {
    cNagraNew n;
    n.Test(data,key,out);
    }
  t=time_ms()-t;
  printf("%d new speed tests: %f s\n",num,(float)t/1000.0);

  return 1;
}
