
#include <stdlib.h>
#include <stdio.h>

#define NAGRA
#define TESTER
#include "crypto.h"
#include "data.h"
#include "system-common.h"
#include "system-nagra.c"

#include "compat.h"

int main(int argc, char *argv[])
{
  if(argc<3) {
    printf("usage: %s <plugin-dir> <id> <rawemm-file>\n",argv[0]);
    return 1;
    }

  InitAll(argv[1]);
  unsigned char emm[64];
  int emmlen=ReadRaw(argv[3],emm,sizeof(emm));
  if(emmlen!=64) {
    printf("bad rawemm file format\n");
    return 1;
    }

  const int romNr=emm[3]&0x0F;
  const int id=strtol(argv[2],0,0);
  printf("Testing with ROM %d, 0xID %04x\n",romNr,id);
  cEmu *emu;
  switch(romNr) {
    case 3:  emu=new cEmuRom3;  break;
    case 7:  emu=new cEmuRom7;  break;
    case 10: emu=new cEmuRom10; break;
    case 11: emu=new cEmuRom11; break;
    default: printf("unsupported ROM\n"); return 1;
    }
  if(!emu || !emu->Init(romNr,id)) {
    printf("initialization failed for ROM %d\n",romNr);
    return 1;
    }

  unsigned char key0[8], key1[8], ki[2];
   if(emu->GetOpKeys(emm,ki,key0,key1)) {
    printf("keyid"); SDump(ki,2);
    printf("key0"); SDump(key0,8);
    printf("key1"); SDump(key1,8);
    }

  delete emu;
}
